

from typing import List

import numpy as np
from deep_sort_reid.metric.IouMetric import IouMetric
from deep_sort_reid.types.detection import Detection


def non_max_suppression(detections: List[List[Detection]],
                        max_overlap: float,
                        confidence_dependent: bool = True
                        ):
    """
    Perform non_max_surpression for boxes of same frame
    with optional confidence dependency
    Return detections, features where indices for every frame has been removed
    """
    detections_to_keep = []

    for frame_idx, detections_frame in enumerate(detections):
        detections_frame_to_keep = []
        detections_frame_idx = []
        if confidence_dependent:
            scores = []
            for detection in detections_frame:
                scores.append(detection.confidence)

            detections_frame_idx = np.flip(np.argsort(scores))

        else:
            for detection_idx, _ in enumerate(detections_frame):
                detections_frame_idx.append(detection_idx)

        for curr_idx, curr_det_idx in enumerate(detections_frame_idx):
            curr_det = detections_frame[curr_det_idx]
            for next_idx in range(len(detections_frame_idx) - 1, curr_idx, -1):
                next_det = detections_frame[detections_frame_idx[next_idx]]

                iou = IouMetric.iou(curr_det.coords, next_det.coords)
                if iou > max_overlap:
                    detections_frame_idx = np.delete(
                        detections_frame_idx, next_idx)

        for det_idx in detections_frame_idx:
            detections_frame_to_keep.append(detections_frame[det_idx])

        detections_to_keep.append(detections_frame_to_keep)

    return detections_to_keep





# KalmanFilter
MEASUREMENT_SPACE_DIM = 4
POS_WEIGHT_DEFAULT = 1/20
VEL_WEIGHT_DEFAULT = 1/160
POS_WEIGHT_FACTOR_INIT = 2
VEL_WEIGHT_FACTOR_INIT = 10
AR_PROJ_STD_DEFAULT = 1e-1
AR_POS_STD_DEFAULT = 1e-2
AR_VEL_STD_DEFAULT = 1e-5
STATE_MEAN_HEIGHT_IDX = 3
"""
GATING_THRESHOLD: 
The x value covering 95% of the area 
of a 4th degree chisquare distirbution
"""
GATING_THRESHOLD = 9.487729036781154





"""
What to pass in:

- features as Tensors or ndarrays, indexed by frames, indexed by objects
- Sequence of Sequence of Detection, indexed as the features

"""


import random
from typing import List
from deep_sort_reid.metric.GatedMetric import GatedMetric
from deep_sort_reid.metric.IouMetric import IouMetric
from deep_sort_reid.storage.CacheStorage import CacheStorage
from deep_sort_reid.models.motion.KalmanFilter import KalmanFilter
from deep_sort_reid.tracker.Tracker import Tracker
from deep_sort_reid.types.tracker import TrackResult
from deep_sort_reid.types.detection import Detection
from deep_sort_reid.types.metric import MetricType
from pydantic import TypeAdapter


class DeepSortReid():

    def __init__(self,
                 # Config
                 metric_type: MetricType = "cosine",
                 features_max_distance: float = 0.5,
                 iou_max_distance: float = 0.5,
                 max_since_update: int = 5,
                 max_samples_per_track: int = 30,
                 hits_until_confirm: int = 3,
                 new_track_max_similarity=0.7,
                 new_track_max_iou=0.5,
                 reid=False,
                 reid_similarity_score=0.9,
                 verbose=False
                 ):
        self.metric_type: MetricType = metric_type
        self.features_max_distance = features_max_distance
        self.iou_max_distance = iou_max_distance
        self.max_samples_per_track = max_samples_per_track
        self.max_since_update = max_since_update
        self.hits_until_confirm = hits_until_confirm
        self.new_track_max_similarity = new_track_max_similarity
        self.new_track_max_iou = new_track_max_iou
        self.reid = reid
        self.reid_similarity_score = reid_similarity_score
        self.verbose = verbose

    def track(self,
              detections: List[List[Detection]]):

        detections = TypeAdapter(
            List[List[Detection]]).validate_python(detections)

        kf = KalmanFilter()
        cache_storage = CacheStorage(self.max_samples_per_track)
        gated_metric = GatedMetric(
            self.metric_type, self.features_max_distance, cache_storage, kf)
        iou_metric = IouMetric('iou', self.iou_max_distance)

        tracker = Tracker(gated_metric, iou_metric,
                          cache_storage, kf,
                          max_since_update=self.max_since_update,
                          hits_until_confirm=self.hits_until_confirm,
                          new_track_max_similarity=self.new_track_max_similarity,
                          new_track_max_iou=self.new_track_max_iou,
                          reid=self.reid,
                          reid_similarity_score=self.reid_similarity_score,
                          )

        no_of_frames = len(detections)
        tracker_results: List[List[TrackResult]] = []
        for _ in range(no_of_frames):
            tracker_results.append([])

        colors = {}

        base_color_bgr = (205, 179, 58)  # BGR values for #3AB3CD

        for frame_idx in range(no_of_frames):
            if self.verbose:
                print("PROCESSEING FRAME: ", frame_idx)

            detections_i = detections[frame_idx]
            tracker.predict()
            tracker.update(detections_i)

            for track in tracker.tracks:
                if track.track_id not in colors:
                    colors[track.track_id] = self.__generate_shade_of_base_color(
                        base_color_bgr)

                track_result = TrackResult(track_id=track.track_id, frame_idx=frame_idx,
                                           coords=track.get_position(), cls=track.cls, state=track.state,
                                           color=colors[track.track_id])

                tracker_results[frame_idx].append(track_result)

        return tracker_results

    def __generate_shade_of_base_color(self, base_color, variation_range=(-30, 30)):
        # Apply variation to the BGR values to generate a lighter or darker shade
        b = min(255, max(0, base_color[0] +
                random.randint(*variation_range)))
        g = min(255, max(0, base_color[1] +
                random.randint(*variation_range)))
        r = min(255, max(0, base_color[2] +
                random.randint(*variation_range)))
        return (b, g, r)





from typing import List, Tuple

from deep_sort_reid.enums.tracker import TrackState
from deep_sort_reid.metric.GatedMetric import GatedMetric
from deep_sort_reid.metric.IouMetric import IouMetric
from deep_sort_reid.storage.CacheStorage import CacheStorage
from deep_sort_reid.models.motion.KalmanFilter import KalmanFilter
from deep_sort_reid.tracker.Track import Track
from deep_sort_reid.types.detection import Detection
from deep_sort_reid.utils.box_methods import from_xyxy_to_xyah
from deep_sort_reid.utils.linear_assignment import min_cost_matching


class Tracker():

    def __init__(self,
                 gated_metric: GatedMetric,
                 iou_metric: IouMetric,
                 cache_storage: CacheStorage,
                 kf: KalmanFilter,
                 max_since_update: int,
                 hits_until_confirm: int,
                 new_track_max_similarity,
                 new_track_max_iou,
                 reid: bool,
                 reid_similarity_score: float,
                 ):

        self.gated_metric = gated_metric
        self.iou_metric = iou_metric
        self.cache_storage = cache_storage
        self.max_since_update = max_since_update
        self.hits_until_confirm = hits_until_confirm
        self.new_track_max_similarity = new_track_max_similarity
        self.new_track_max_iou = new_track_max_iou
        self.reid = reid
        self.reid_similarity_score = reid_similarity_score

        self.kf = kf
        self.tracks: List[Track] = []
        self.deleted_tracks: List[Track] = []
        self.next_tracker_id = 1

    def __initiate_track(self, detection: Detection, track_id: int = -1):
        new_track_id = self.next_tracker_id
        if track_id != -1:
            new_track_id = track_id

        state_mean, state_covariance = self.kf.initiate(
            from_xyxy_to_xyah(detection.coords))
        new_track = Track(
            state_mean,
            state_covariance,
            new_track_id,
            self.hits_until_confirm,
            self.max_since_update,
            detection.feature,
            detection.cls,
        )

        self.tracks.append(new_track)
        self.next_tracker_id += 1

    def predict(self):
        for track in self.tracks:
            track.predict(self.kf)

    def update(self, detections_i: List[Detection]):
        confirmed_tracks_idx = []
        unconfirmed_tracks_idx = []
        deleted_tracks_idx = []

        for track_idx, track in enumerate(self.tracks):
            if track.state == TrackState.CONFIRMED:
                confirmed_tracks_idx.append(track_idx)
            elif track.state == TrackState.UNCONFIRMED:
                unconfirmed_tracks_idx.append(track_idx)
            else:
                deleted_tracks_idx.append(track_idx)

        matches_track_det_idx, \
            unmatched_tracks_idx, \
            unmatched_detections_idx = \
            self.__match(detections_i, confirmed_tracks_idx,
                         unconfirmed_tracks_idx)

        # Update the state of the matched tracks with new measurement information
        for track_idx, detection_idx in matches_track_det_idx:
            self.tracks[track_idx].update(
                self.kf,
                detections_i[detection_idx])

        for detection_idx in unmatched_detections_idx:
            detection = detections_i[detection_idx]
            if self.reid:
                self.__match_reid(detection, detection_idx,
                                  matches_track_det_idx)

            # We use the __iou_detection_tracks_check to avoid initating a new track
            # in a position where another one already resides
            elif self.__iou_detection_tracks_check(detection):
                self.__initiate_track(detection)

        # Potentially remove tracks
        for track_idx in unmatched_tracks_idx:
            self.tracks[track_idx].no_match()

        for track_idx, detection_idx in matches_track_det_idx:
            track = self.tracks[track_idx]
            detection = detections_i[detection_idx]

            if detection.feature is not None:
                self.cache_storage.add_sample(
                    track.track_id, detection.feature)

        tracks_to_keep: List[Track] = []
        for track in self.tracks:
            if (track.state != TrackState.DELETED):
                tracks_to_keep.append(track)
            else:
                self.deleted_tracks.append(track)

        self.tracks = tracks_to_keep

    def __match_reid(self, detection: Detection, detection_idx: int, matches_track_det_idx: List[Tuple]):
        # Potentially re-assign in-frame or deleted tracks based on visual features
        score, track_idx = self.__similarity_detection_tracks_match(
            detection, self.tracks)
        deleted_score, deleted_track_idx = self.__similarity_detection_tracks_match(
            detection, self.deleted_tracks)

        if score > self.reid_similarity_score or deleted_score > self.reid_similarity_score:
            if score > deleted_score:
                matches_tracks_idx = [
                    match_track_idx for match_track_idx, _ in matches_track_det_idx]

                if track_idx not in matches_tracks_idx:
                    self.tracks[track_idx].update(
                        self.kf,
                        detection)
                    matches_track_det_idx.append(
                        (track_idx, detection_idx))

                return

            else:
                re_assigned_track_id = self.deleted_tracks[deleted_track_idx].track_id
                self.__initiate_track(
                    detection, re_assigned_track_id)
                re_assigned_track_idx = len(self.tracks) - 1
                matches_track_det_idx.append(
                    (re_assigned_track_idx, detection_idx))

                del self.deleted_tracks[deleted_track_idx]

        elif self.__iou_detection_tracks_check(detection):
            self.__initiate_track(detection)

        elif self.__iou_detection_tracks_check(detection):
            self.__initiate_track(detection)

    def __similarity_detection_tracks_match(self, detection: Detection, tracks: List[Track]):
        highest_similarity_score = 0
        highest_similarity_track_idx = -1

        for track_idx, track in enumerate(tracks):
            track_samples = None

            if track.track_id in self.cache_storage.samples:
                track_samples = self.cache_storage.get(track.track_id, 'mean')

            if (track_samples is not None and detection.feature is not None):
                # 1 - in order to transform it back to similarity from distance
                # Should be setup to work with euclidean distance likewise
                similarity_score = (1 - GatedMetric.cosine_distance(
                    track_samples, [detection.feature])).item()

                if similarity_score > highest_similarity_score:
                    highest_similarity_score = similarity_score
                    highest_similarity_track_idx = track_idx

        return highest_similarity_score, highest_similarity_track_idx

    def __iou_detection_tracks_check(self, detection: Detection):
        highest_iou = 0

        tracks_to_check = self.tracks.copy()

        if self.reid:
            for track in self.deleted_tracks:
                if track.time_since_update < 3:
                    tracks_to_check.append(track)

        for track in tracks_to_check:
            iou = IouMetric.iou(detection.coords, track.get_position())

            if iou > highest_iou:
                highest_iou = iou
        if highest_iou >= self.new_track_max_iou:
            return False

        return True

    def __match(self, detections_i: List[Detection], confirmed_tracks_idx: List[int], unconfirmed_tracks_idx: List[int]):
        matches_features_track_det_idx, \
            unmatched_features_tracks_idx, \
            unmatched_features_detections_idx = \
            self.__match_features_max_since_update(detections_i,
                                                   confirmed_tracks_idx)

        unmatched_tracks_idx_age_one = []
        unmatched_tracks_idx_older = []

        for track_idx in unmatched_features_tracks_idx:
            if self.tracks[track_idx].time_since_update > 5:
                unmatched_tracks_idx_older.append(track_idx)
            else:
                unmatched_tracks_idx_age_one.append(track_idx)

        iou_track_candidates = unconfirmed_tracks_idx + unmatched_tracks_idx_age_one
        unmatched_features_tracks_idx = unmatched_tracks_idx_older

        matches_iou_track_det_idx, \
            unmatched_iou_tracks_idx, \
            unmatched_iou_detections_idx = \
            self.__match_iou(detections_i, iou_track_candidates,
                             unmatched_features_detections_idx)

        matches_track_det_idx = matches_features_track_det_idx + matches_iou_track_det_idx
        unmatched_tracks_idx: List[int] = list(
            set(unmatched_features_tracks_idx + unmatched_iou_tracks_idx))
        unmatched_detections_idx = unmatched_iou_detections_idx

        return matches_track_det_idx, unmatched_tracks_idx, unmatched_detections_idx

    def __match_features_max_since_update(self,
                                          detections: List[Detection],
                                          tracks_idx: List[int]):
        """
        Matches detections from the current frame (i) to existing tracks that have been updated within
        the `max_since_update` range. This method is primarily for re-matching objects that are in-frame.
        """

        # The detections_idx we haven't matched yet
        detections_idx = []

        for detection_idx in range(len(detections)):
            detections_idx.append(detection_idx)

        matches_track_det_idx: List[Tuple[int, int]] = []
        unmatched_detections_idx: List[int] = detections_idx

        # Note that we start from tracks of age 1 (frame)
        for age in range(1, self.max_since_update):
            if len(detections_idx) == 0:
                break

            tracks_idx_curr_age = []
            for track_idx in tracks_idx:
                if self.tracks[track_idx].time_since_update == age:
                    tracks_idx_curr_age.append(track_idx)

            if len(tracks_idx_curr_age) == 0:
                continue

            matches_track_det_idx_curr_age, \
                _, \
                unmatched_detections_idx = min_cost_matching(self.gated_metric,
                                                             self.tracks,
                                                             detections,
                                                             tracks_idx_curr_age,
                                                             unmatched_detections_idx)

            matches_track_det_idx.extend(matches_track_det_idx_curr_age)

        # We need to make sure to return all tracks_idx that weren't matched
        unmatched_tracks_idx = list(set(
            tracks_idx) - set(matched_track_idx for matched_track_idx, _ in matches_track_det_idx))

        return matches_track_det_idx, unmatched_tracks_idx, unmatched_detections_idx

    def __match_iou(self, detections: List[Detection], tracks_idx: List[int], detections_idx: List[int]):
        matches_track_det_idx, \
            unmatched_tracks_idx, \
            unmatched_detections_idx = \
            min_cost_matching(self.iou_metric, self.tracks,
                              detections, tracks_idx, detections_idx)

        return matches_track_det_idx, unmatched_tracks_idx, unmatched_detections_idx





from typing import List

from ultralytics import YOLO

from deep_sort_reid.types.coords import CoordinatesXYXY
from deep_sort_reid.types.detection import Detection
from deep_sort_reid.utils.misc import get_device


def detect_objects_yolo(file_input_path: str, model: YOLO, model_params={
                        "stream": True,
                        "classes": [0],
                        "iou": 0.5,
                        "save": False,
                        "conf": 0.5,
                        "device": get_device()
                        }) -> List[List[Detection]]:
    """
    A helper function for YOLO object detection

    Returns 2d list of detections, where first dimension is frames and second is detection
    """

    results = model(file_input_path, **model_params)

    frames_detections: List[List[Detection]] = []
    for frame_objects in results:

        detections: List[Detection] = []
        for object in frame_objects:

            startX, startY, endX, endY = object.boxes.xyxy[0].tolist()
            coords = CoordinatesXYXY(
                start_x=int(startX),
                start_y=int(startY),
                end_x=int(endX),
                end_y=int(endY))

            detection_model = Detection(
                coords=coords, cls=[int(object.boxes.cls.item())],
                confidence=object.boxes.conf, feature=None)

            detections.append(detection_model)
        frames_detections.append(detections)

    return frames_detections





"""
From to box formats
"""
from torch import Tensor
from deep_sort_reid.types.coords import CoordinatesXYAH
from deep_sort_reid.types.coords import CoordinatesXYXY


def from_xyah_to_tensor(box: CoordinatesXYAH) -> Tensor:
    return Tensor([box.center_x, box.center_y, box.aspect_ratio, box.height])


def from_xyxy_to_xyah(box: CoordinatesXYXY) -> CoordinatesXYAH:

    right_x = max(box.start_x, box.end_x)
    left_x = min(box.start_x, box.end_x)
    top_y = min(box.start_y, box.end_y)
    bottom_y = max(box.start_y, box.end_y)

    width = right_x - left_x
    height = bottom_y - top_y
    center_x = left_x + width / 2
    center_y = top_y + height / 2
    aspect_ratio = width / height

    return CoordinatesXYAH(center_x=center_x,
                           center_y=center_y,
                           aspect_ratio=aspect_ratio,
                           height=height)



from typing import List, Sequence

from torch import Tensor
import torch
from deep_sort_reid.tracker.Track import Track
from deep_sort_reid.types.detection import Detection
from deep_sort_reid.types.metric import MetricType
from abc import ABC, abstractmethod


class Metric():

    def __init__(self, metric_type: MetricType, max_distance: float):
        self.metric_type: MetricType = metric_type
        self.max_distance = max_distance

    @abstractmethod
    def distance(self, tracks: List[Track],
                 detections: List[Detection]) -> torch.Tensor:
        """
        Create a cost matrix
        where rows indicate detection targets
        amd columns the features.
        For every row (detection target), calculate
        the distance using the chosen metric, between the 
        current row detection features and the rest of the provided 
        features   
        """

        return torch.Tensor()

    @abstractmethod
    def __call__(self,
                 tracks: List[Track],
                 detections: List[Detection],
                 ) -> Tensor:

        return torch.Tensor()





from typing import List

from torch import Tensor
import torch
from deep_sort_reid.tracker.Track import Track
from deep_sort_reid.types.coords import CoordinatesXYXY
from deep_sort_reid.types.detection import Detection
from deep_sort_reid.types.metric import MetricType
from deep_sort_reid.metric.Metric import Metric


class IouMetric(Metric):

    def __init__(self, metric_type: MetricType, max_distance: float):
        self.metric_type: MetricType = metric_type
        self.max_distance = max_distance

    def __distance(self,
                   tracks: List[Track],
                   detections: List[Detection],
                   ) -> Tensor:

        cost_matrix = torch.zeros((len(tracks)), len(detections))

        for idx, track in enumerate(tracks):
            if track.time_since_update > 5:
                cost_matrix[idx, :] = torch.inf
                continue

            track_state_coords: CoordinatesXYXY = track.get_position()

            cost_matrix[idx, :] = self.__iou_distances(
                track_state_coords, detections)

        return cost_matrix

    def __iou_distances(self, track_state_coords: CoordinatesXYXY, detections: List[Detection]):

        distances = []

        for detection in detections:
            iou_distance = 1 - self.iou(track_state_coords, detection.coords)
            distances.append(iou_distance)

        return torch.Tensor(distances)

    @staticmethod
    def iou(box_1: CoordinatesXYXY, box_2: CoordinatesXYXY):
        (box_1_start_x, box_1_end_x,
         box_1_start_y, box_1_end_y) = (min(box_1.start_x, box_1.end_x), max(box_1.start_x, box_1.end_x),
                                        min(box_1.start_y, box_1.end_y), max(box_1.start_y, box_1.end_y))

        (box_2_start_x, box_2_end_x,
         box_2_start_y, box_2_end_y) = (min(box_2.start_x, box_2.end_x), max(box_2.start_x, box_2.end_x),
                                        min(box_2.start_y, box_2.end_y), max(box_2.start_y, box_2.end_y))

        intersection_width = max(0, min(
            box_1_end_x, box_2_end_x) - max(box_1_start_x, box_2_start_x))

        intersection_height = max(0, min(
            box_1_end_y, box_2_end_y) - max(box_1_start_y, box_2_start_y))

        intersection_area = intersection_width * intersection_height

        box_1_area = (box_1_end_x - box_1_start_x) * \
            (box_1_end_y - box_1_start_y)

        box_2_area = (box_2_end_x - box_2_start_x) * \
            (box_2_end_y - box_2_start_y)

        union_area = box_1_area + box_2_area - intersection_area

        iou = (intersection_area / union_area)
        return iou

    def __call__(self,
                 tracks: List[Track],
                 detections: List[Detection],
                 ):

        return self.__distance(tracks, detections)
